import pkg_resources

from anonlink import bloommatcher
from anonlink import entitymatch
from anonlink import network_flow

__version__ = pkg_resources.get_distribution('anonlink').version
__author__ = 'N1 Analytics'
